from datetime import date
import os
import sys

code_dir = os.path.join(os.path.abspath("."), "src")
os.chdir(code_dir)
sys.path.append(code_dir)

from datetime import date
import pandas as pd

# import classes
# import classes
import datetime
from markets_insights.datareader import data_reader
from markets_insights.dataprocess import data_processor

reader = data_reader.MemoryCachedDataReader(data_reader.BhavCopyReader())
criteria = data_reader.DateRangeCriteria(from_date=date(2024, 1, 1), to_date=date(2024, 1, 31))
reader.read(criteria)

# Fetch the data
# reader = data_reader.BhavCopyReader()
# options = data_processor.HistoricalDataProcessOptions(include_monthly_data = False, include_annual_data=False)
# histDataProcessor = data_processor.HistoricalDataProcessor(options)

# from_date = datetime.date(2023, 12, 1)
# to_date = datetime.date(2023, 12, 31)
# result = histDataProcessor.process(data_reader.BhavCopyReader(), data_reader.DateRangeCriteria(from_date, to_date))


from datetime import date
import os
import sys

code_dir = os.path.join(os.path.abspath("."), "src")
os.chdir(code_dir)
sys.path.append(code_dir)

import datetime
from markets_insights.datareader.data_reader import DateRangeCriteria, NseIndicesReader
from markets_insights.dataprocess.data_processor import HistoricalDataProcessor, HistoricalDataProcessOptions

from markets_insights.datareader.data_reader import NseIndicesReader, DateRangeDataReader, DateRangeCriteria
from markets_insights.core.core import IdentifierFilter

nifty50_reader = NseIndicesReader().set_filter(IdentifierFilter("NIFTY 50"))
indiavix_reader = NseIndicesReader().set_filter(IdentifierFilter("India VIX"))
reader = nifty50_reader / indiavix_reader

histDataProcessor = HistoricalDataProcessor()

# Fetch the data
from_date = datetime.date(2024, 1, 1)
to_date = datetime.date(2024, 1, 31)
result = histDataProcessor.process(reader, DateRangeCriteria(from_date, to_date))


import math

from attr import dataclass
import markets_insights as mi

from urllib.request import urlopen
from zipfile import ZipFile
from markets_insights.core.column_definition import (
    BaseColumns,
    BasePriceColumns,
    DerivativesBaseColumns,
)
from markets_insights.core.environment import EnvironmentSettings
import os
from string import Template
import pandas as pd
from urllib.error import HTTPError
from datetime import date, timedelta

from markets_insights.core.core import (
    FilterBase,
    InstrumentTypeFilter,
    MarketDaysHelper,
    Instrumentation,
    TypeHelper,
)

from enum import Enum
class Status(Enum):
    NONE: int = 0
    PARTIAL: int = 1
    COMPLETE: int = 2
    UKNOWN: int = 3


class ReaderDateCriteria:
    pass

@dataclass
class ForDateCriteria(ReaderDateCriteria):
    for_date: date


@dataclass
class DateRangeCriteria(ReaderDateCriteria):
    from_date: date
    to_date: date


@dataclass
class MultiDatesCriteria(ReaderDateCriteria):
    for_dates: list[date]


@dataclass(init=False)
class ReaderDataAvailibility:
    from_date: date = None
    till_date: date = None

    def __init__(self, from_date: date = date(1900, 1, 1), till_date: date = date.today()):
        self.from_date = from_date
        self.till_date = till_date


@dataclass
class ReaderDataAvailibilityStatus:
    status: int
    availibility_ranges: list[DateRangeCriteria] = None
    unavailibility_ranges: list[DateRangeCriteria] = None


class ReaderOptions:
    filename = Template("")
    url_template = Template("")
    output_path_template = Template("")
    unzip_path_template = Template("")
    primary_data_path_template = Template("")
    unzip_file: bool = True
    data_availibility: ReaderDataAvailibility = None
    download_timeout = 5  # seconds
    col_prefix = None

@dataclass
class ReaderRescaleOptions:
    volume_scale: int = 1
    turnover_scale: int = 1

def get_safe_min_date(for_date) -> date:
    return for_date if for_date is not None else date(1900, 1, 1)


def get_safe_max_date(for_date) -> date:
    return for_date if for_date is not None else date.today()


class DataReader:
    options: ReaderOptions

    def __init__(self):
        self.options: ReaderOptions = ReaderOptions()
        self.rescale_options: ReaderRescaleOptions = ReaderRescaleOptions()
        self.name: str = ""
        self.filter: FilterBase = None

    def date_is_between_availibility(self, for_date) -> bool:
        return get_safe_min_date(self.options.data_availibility.from_date) <= for_date <= get_safe_max_date(self.options.data_availibility.till_date)

    def has_data(self, criteria: ReaderDateCriteria) -> ReaderDataAvailibilityStatus:
        if self.options.data_availibility:
            if isinstance(criteria, ForDateCriteria):
                if self.date_is_between_availibility(criteria.for_date):
                    return ReaderDataAvailibilityStatus(status=Status.COMPLETE)
                else:
                    return ReaderDataAvailibilityStatus(status=Status.NONE)
            elif isinstance(criteria, DateRangeCriteria):
                availibility_status = ReaderDataAvailibilityStatus(status=Status.NONE)
                availibility_status.unavailibility_ranges = []
                from_date, to_date = MarketDaysHelper.get_this_or_next_market_day(criteria.from_date), MarketDaysHelper.get_this_or_previous_market_day(criteria.to_date)

                if (self.options.data_availibility.from_date > from_date and to_date < self.options.data_availibility.from_date) \
                        or (from_date > self.options.data_availibility.till_date):
                    availibility_status.status = Status.NONE
                    availibility_status.unavailibility_ranges.append(criteria)
                elif self.options.data_availibility.from_date <= from_date and self.options.data_availibility.till_date >= to_date:
                    availibility_status.status = Status.COMPLETE
                else:
                    if self.options.data_availibility.from_date > from_date:
                        availibility_status.unavailibility_ranges.append(DateRangeCriteria(from_date, self.options.data_availibility.from_date - timedelta(days=1)))
                    if self.options.data_availibility.till_date < to_date:
                        availibility_status.unavailibility_ranges.append(DateRangeCriteria(self.options.data_availibility.till_date + timedelta(days=1), to_date))
                    availibility_status.status = Status.PARTIAL
                    availibility_status.availibility_ranges = [
                            DateRangeCriteria(
                                max([self.options.data_availibility.from_date, from_date]),
                                min([self.options.data_availibility.till_date, to_date])
                        )]
                return availibility_status
            else:
                return ReaderDataAvailibilityStatus(status=Status.UKNOWN)
        else:
            return ReaderDataAvailibilityStatus(status=Status.UKNOWN)
        

    def set_filter(self, filter: FilterBase):
        self.filter = filter
        return self

    def get_date_parts(self, for_date: date):
        return {
            "year": str(for_date.year),
            "month": for_date.strftime("%B").upper()[:3],
            "day": str(for_date.strftime("%d")),
        }

    def merge_with_operation_on_price(
        self, data_l: pd.DataFrame, data_r: pd.DataFrame, operation
    ) -> pd.DataFrame:
        merged_df = pd.merge(
            data_l,
            data_r,
            how="inner",
            on=[BaseColumns.Identifier, BaseColumns.Date],
            left_index=False,
            right_index=False,
        )

        for col in TypeHelper.get_class_static_values(BasePriceColumns):
            merged_df[col] = operation(
                merged_df(merged_df[col + "_x"], merged_df[col + "_y"])
            )

        return merged_df

    def read(self, criteria: ReaderDateCriteria) -> pd.DataFrame:
        data = self.read_data(criteria.for_date)
        return self.post_read_data(data)
    
    def post_read_data(self, data: pd.DataFrame) -> pd.DataFrame:
        if not (data is None or data.empty):
            column_name_mappings = self.get_column_name_mappings()
            if column_name_mappings is not None:
                data.rename(columns=column_name_mappings, inplace=True)

            data.drop_duplicates(inplace=True)

            self.normalise_base_column_values(data)
            
            if self.filter:
                data = data.query(str(self.filter))

            return self.sanitize_data(data)
        else:
            return data

    def read_data(self, for_date: date) -> pd.DataFrame:
        date_parts = self.get_date_parts(for_date)
        filenames = self.get_filenames(for_date)
        output_file_path = self.options.output_path_template.substitute(
            **({**EnvironmentSettings.Paths, **filenames})
        )
        primary_data_file_path = self.options.primary_data_path_template.substitute(
            **({**EnvironmentSettings.Paths, **filenames})
        )

        if not os.path.exists(output_file_path):
            Instrumentation.debug(f"Downloading data for {for_date}")
            url = self.options.url_template.substitute(**({**date_parts, **filenames}))

            Instrumentation.debug(url)

            urldata = urlopen(url, timeout=self.options.download_timeout)
            with open(output_file_path, "wb") as output:
                output.write(urldata.read())

        unzip_folder_path = self.options.unzip_path_template.substitute(
            **({**EnvironmentSettings.Paths, **filenames})
        )
        if self.options.unzip_file == True and not os.path.exists(unzip_folder_path):
            self.unzip_content(output_file_path, unzip_folder_path, for_date)

        data = self.read_data_from_file(for_date, primary_data_file_path)

        return data

    def normalise_base_column_values(self, data: pd.DataFrame) -> pd.DataFrame:
        for col_name in [BaseColumns.Open, BaseColumns.High, BaseColumns.Low]:
            data[col_name] = data.apply(
                lambda x: x[col_name]
                if str(x[col_name]).replace(".", "").isnumeric() == True
                else x[BaseColumns.Close],
                axis=1,
            )
            data[col_name] = data[col_name].astype(float)

        data[BaseColumns.PreviousClose] = data.groupby(
            BaseColumns.Identifier
        )[BaseColumns.Close].transform(lambda x: x.shift(1))

        for col_name in [BaseColumns.Volume, BaseColumns.Turnover]:
            if col_name in data.columns:
                data[col_name] = (
                    data[col_name]
                    .apply(
                        lambda val: val
                        if str(val).replace(".", "").isnumeric() == True
                        else math.nan
                    )
                    .astype(float)
                )
            else:
                data[col_name] = 0

        for rescaling in [
            (BaseColumns.Volume, self.rescale_options.volume_scale),
            (BaseColumns.Turnover, self.rescale_options.turnover_scale),
        ]:
            if rescaling[1] != 1:
                data[rescaling[0]] = data[rescaling[0]] * rescaling[1]

        return data

    def unzip_content(self, output_file_path, unzip_folder_path, for_date):
        zf = ZipFile(output_file_path)
        zf.extractall(path=unzip_folder_path)
        zf.close()

    def read_data_from_file(self, for_date, primary_data_filepath):
        primary_data = pd.read_csv(primary_data_filepath)
        primary_data["Date"] = pd.to_datetime(for_date)
        return primary_data

    def get_filenames(self, for_date):
        return Exception("Not implemented!")

    def get_column_name_mappings(self):
        return None

    def sanitize_data(self, data):
        return data

    def __sub__(self, other):
        return ArithmaticOpReader(
            left=self, right=other, operator=lambda x, y: x - y, op_symbol="-"
        )

    def __add__(self, other):
        return ArithmaticOpReader(
            left=self, right=other, operator=lambda x, y: x + y, op_symbol="+"
        )

    def __mul__(self, other):
        return ArithmaticOpReader(
            left=self, right=other, operator=lambda x, y: x * y, op_symbol="*"
        )

    def __truediv__(self, other):
        return ArithmaticOpReader(
            left=self, right=other, operator=lambda x, y: x / y, op_symbol="/"
        )


def get_date_criteria_based_reader(reader: DataReader, criteria: ReaderDateCriteria) -> DataReader:
    if isinstance(criteria, DateRangeCriteria) and not isinstance(reader, DateRangeSourceDataReader):
        return DateRangeDataReaderWrapper(reader)
    elif isinstance(criteria, MultiDatesCriteria) and not isinstance(reader, MultiDatesDataReader):
        return MultiDatesDataReader(reader)
    else:
        return reader
    

class MultiDatesDataReader(DataReader):
    reader: DataReader

    def __init__(self, reader: DataReader):
        super().__init__()
        self.reader = reader
        self.name = reader.name
        self.filter = reader.filter
        self.options = reader.options

    def read(self, criteria: ReaderDateCriteria):
        if not isinstance(criteria, MultiDatesCriteria):
            raise Exception("MultiDatesDataReader.read() expects MultiDatesCriteria")
        
        result = pd.DataFrame()
        for for_date in criteria.for_dates:
            if MarketDaysHelper.is_open_for_day(pd.Timestamp(for_date).date()):
                try:
                    data = self.reader.read(ForDateCriteria(for_date))
                    
                    if result.empty:
                        result = data
                    else:
                        if not data.empty:
                            result = pd.concat(
                                [result, data], ignore_index=True
                            ).reset_index(drop=True)
                except Exception as e:
                    print(e, for_date.strftime("date(%Y, %m, %d),"))

        return self.post_read_data(result)


class SingleDaySourceDataReader(DataReader):
    pass


class DateRangeSourceDataReader(DataReader):
    pass

class DateRangeDataReaderWrapper(DateRangeSourceDataReader):
    reader: DataReader

    def __init__(self, reader: DataReader):
        super().__init__()
        if reader:
            self.name = reader.name
            self.filter = reader.filter
            self.options = reader.options
            self.reader = reader

    def read(self, criteria: ReaderDateCriteria):
        if not isinstance(criteria, ReaderDateCriteria):
            raise Exception("DateRangeDataReader.read() expects ReaderDateCriteria")
        
        datelist = MarketDaysHelper.get_days_list_for_range(criteria.from_date, criteria.to_date)
        result = pd.DataFrame()
        for for_date in datelist:
            if MarketDaysHelper.is_open_for_day(pd.Timestamp(for_date).date()):
                try:
                    data = self.reader.read(ForDateCriteria(for_date))
                    
                    if result.empty:
                        result = data
                    else:
                        if not data.empty:
                            result = pd.concat(
                                [result, data], ignore_index=True
                            ).reset_index(drop=True)
                except Exception as e:
                    print(e, for_date.strftime("date(%Y, %m, %d),"))

        return self.post_read_data(result)

    def has_data(self, criteria: ReaderDateCriteria):
        if self.reader:
            return self.reader.has_data(criteria)
        else:
            return self.has_data()


class ChainedDataReader(DateRangeSourceDataReader):
    def __init__(self, next: DataReader):
        super().__init__()
        if not isinstance(next, DateRangeSourceDataReader):
            self.next = DateRangeDataReaderWrapper(next)
        else:
            self.next = next
    
    def read(self, criteria: ReaderDateCriteria) -> pd.DataFrame:
        # check has data for date range
        availibility: ReaderDataAvailibilityStatus = self.has_data(criteria)
        
        if availibility.status == Status.COMPLETE:
            data = self.read_data(criteria)
        elif availibility.status == Status.NONE or availibility.status == Status.UKNOWN:
            data = self.next.read(criteria)
            if not data.empty:
                #data = self.post_read_data(data)
                self.on_received_more_data(data)
        elif availibility.status == Status.PARTIAL:
            available_data = self.read_data(criteria)
            unavailable_data: list[pd.DataFrame] = []
            
            for date_range in availibility.unavailibility_ranges:
                data = self.next.read(date_range)
                if not data.empty:
                    unavailable_data.append(data)
            if len(unavailable_data) > 0:
                all_unavailable_data = pd.concat(unavailable_data).sort_values([BaseColumns.Date])
                #all_unavailable_data = self.post_read_data(all_unavailable_data)
                data = pd.concat([all_unavailable_data, available_data])
                data[BaseColumns.Date] = data[BaseColumns.Date].astype('datetime64[ns]')
                data = data.sort_values([BaseColumns.Date])
                self.on_received_more_data(all_unavailable_data)
            else:
                data = available_data
        
        return self.post_read_data(data)
    
    def on_received_more_data(self, data: list[pd.DataFrame]):
        pass


class CachedDataReader(ChainedDataReader):
    def __init__(self, next: DataReader):
        super().__init__(next)
        self.name = next.name
        self.options.col_prefix = next.options.col_prefix
    
    def read_data(self, criteria) -> pd.DataFrame:
        return self.read_cached_data(criteria)
    
    def read_cached_data(self, criteria) -> pd.DataFrame:
        raise "Not implemented exception"
    
    def post_read_data(self, data: pd.DataFrame) -> pd.DataFrame:
        data = data[data.columns.drop(list(data.filter(regex='Unnamed')))]
        data = data.reset_index(drop=True)
        data = data.drop_duplicates()
        if self.filter:
            data = data.query(str(self.filter))
        return data


class MemoryCachedDataReader(CachedDataReader):
    def __init__(self, next: DataReader):
        super().__init__(next)
        self.cached_data = pd.DataFrame()
    
    def read_cached_data(self, criteria: DateRangeCriteria) -> pd.DataFrame:
        return self.cached_data[
            self.cached_data[BaseColumns.Date].dt.date.between(criteria.from_date, criteria.to_date)
        ]
        
    def on_received_more_data(self, new_data: pd.DataFrame):
        existing_data = self.cached_data
        self.cached_data = pd.concat([existing_data, new_data])
        
        self.cached_data.sort_values(BaseColumns.Date)
        if self.options.data_availibility is None:
            self.options.data_availibility = ReaderDataAvailibility()
        self.options.data_availibility.from_date = self.cached_data[BaseColumns.Date].min().date()
        self.options.data_availibility.till_date = self.cached_data[BaseColumns.Date].max().date()


class ArithmaticOpReader(DataReader):
    def __init__(self, left: DataReader, right: DataReader, operator, op_symbol: str):
        super().__init__()
        self.l_reader = left
        self.r_reader = right
        self.operator = operator
        self.op_symbol = op_symbol
        self.l_prefix = None
        self.r_prefix = None
        self.options.col_prefix = ""
        self.name = f"{left.name}{op_symbol}{right.name}"

    def read(self, criteria: ReaderDateCriteria) -> pd.DataFrame:
        l_data = get_date_criteria_based_reader(self.l_reader, criteria).read(criteria)
        r_data = get_date_criteria_based_reader(self.r_reader, criteria).read(criteria)
        if not (l_data.empty or r_data.empty):
            on_cols = [BaseColumns.Identifier, BaseColumns.Date]
            prefix_l = self.l_reader.options.col_prefix
            prefix_r = self.r_reader.options.col_prefix

            l_data_unique_id_count = len(l_data[BaseColumns.Identifier].unique())
            r_data_unique_id_count = len(r_data[BaseColumns.Identifier].unique())
            
            if l_data_unique_id_count == 1:
                on_cols = [BaseColumns.Date]
                prefix_l = l_data[BaseColumns.Identifier].values[0] + "-"
            
            if r_data_unique_id_count == 1:
                on_cols = [BaseColumns.Date]
                prefix_r = r_data[BaseColumns.Identifier].values[0] + "-"
            
            if not (
                f"{prefix_l}{BaseColumns.Close}" in l_data.columns
                and f"{prefix_r}{BaseColumns.Close}" in l_data.columns
            ):
                merged_df = pd.merge(
                    l_data,
                    r_data,
                    how="inner",
                    on=on_cols,
                    left_index=False,
                    right_index=False,
                )

                if BaseColumns.Identifier not in on_cols:
                    merged_df[BaseColumns.Identifier] = merged_df[BaseColumns.Identifier + "_x"] + " " + self.op_symbol + " " + merged_df[BaseColumns.Identifier + "_y"]

                update_col_prefix = {}
                for col in [col for col in merged_df.columns if "_x" in col]:
                    update_col_prefix[col] = f"{prefix_l}{col.replace('_x', '')}"
                for col in [col for col in merged_df.columns if "_y" in col]:
                    update_col_prefix[col] = f"{prefix_r}{col.replace('_y', '')}"

                merged_df.rename(columns=update_col_prefix, inplace=True)
            else:
                merged_df = l_data

            for col in TypeHelper.get_class_static_values(BasePriceColumns):
                merged_df[col] = self.operator(
                    merged_df[f"{prefix_l}{col}"], merged_df[f"{prefix_r}{col}"]
                )

            self.l_prefix = prefix_l
            self.r_prefix = prefix_r

            return merged_df
        else:
            return pd.DataFrame()


class BhavCopyReader(SingleDaySourceDataReader):
    def __init__(self):
        super().__init__()
        self.name = "nse_equities"
        self.options.col_prefix = "Cash-"
        __base_url = "https://archives.nseindia.com/content/historical/EQUITIES/"
        self.options.data_availibility = ReaderDataAvailibility(from_date=date.fromisoformat("2016-01-01"))
        self.options.url_template = Template(
            __base_url + "$year/$month/$download_filename"
        )
        self.options.output_path_template = Template(
            f"$DataBaseDir/$RawDataDir/$BhavDataDir/$download_filename"
        )
        self.options.unzip_path_template = Template(
            "$DataBaseDir/$RawDataDir/$BhavDataDir/$download_filename_wo_ext"
        )
        self.options.primary_data_path_template = Template(
            "$DataBaseDir/$RawDataDir/$BhavDataDir/$download_filename_wo_ext/$primary_data_filename"
        )

    def get_filenames(self, for_date):
        __formatted_date = for_date.strftime("%d%b%Y").upper()
        return {
            "download_filename_wo_ext": f"cm{__formatted_date}bhav",
            "download_filename": f"cm{__formatted_date}bhav.csv.zip",
            "primary_data_filename": f"cm{__formatted_date}bhav.csv",
        }

    def get_column_name_mappings(self):
        return {
            "SYMBOL": BaseColumns.Identifier,
            "PREVCLOSE": BaseColumns.PreviousClose,
            "OPEN": BaseColumns.Open,
            "HIGH": BaseColumns.High,
            "LOW": BaseColumns.Low,
            "CLOSE": BaseColumns.Close,
            "TOTTRDQTY": BaseColumns.Volume,
            "TOTTRDVAL": BaseColumns.Turnover,
        }

    def sanitize_data(self, data):
        return data[data["SERIES"] == "EQ"].reset_index(drop=True)



class NseIndicesNewReader(SingleDaySourceDataReader):
    def __init__(self):
        super().__init__()
        self.name = "nse_indices"
        self.options.col_prefix = "index-"
        self.rescale_options.turnover_scale = math.pow(10, 7)
        __base_url = "https://archives.nseindia.com/content/indices/"
        self.options.unzip_file = False
        self.options.url_template = Template(__base_url + "$download_filename")
        self.options.output_path_template = Template(
            "$DataBaseDir/$RawDataDir/$NseIndicesDataDir/$download_filename"
        )
        self.options.primary_data_path_template = Template(
            "$DataBaseDir/$RawDataDir/$NseIndicesDataDir/$download_filename"
        )
        self.options.data_availibility = ReaderDataAvailibility(from_date=date.fromisoformat("2013-01-01"))

    def get_filenames(self, for_date: date):
        __formatted_date = for_date.strftime("%d%m%Y").upper()
        return {
            "download_filename": f"ind_close_all_{__formatted_date}.csv",
            "primary_data_filename": f"ind_close_all_{__formatted_date}.csv",
        }

    def get_column_name_mappings(self):
        return {
            "Index Name": BaseColumns.Identifier,
            "Open Index Value": BaseColumns.Open,
            "High Index Value": BaseColumns.High,
            "Low Index Value": BaseColumns.Low,
            "Closing Index Value": BaseColumns.Close,
            "Volume": BaseColumns.Volume,
            "Turnover (Rs. Cr.)": BaseColumns.Turnover,
        }


class NseDerivatiesReaderBase(SingleDaySourceDataReader):
    def __init__(self):
        super().__init__()
        self.options.data_availibility = ReaderDataAvailibility(from_date=date(2016, 1, 1))

    def sanitize_data(self, data):
        return data[data["OpenInterest"] > 0].reset_index(drop=True)


class NseDerivatiesReader(NseDerivatiesReaderBase):
    def __init__(self):
        super().__init__()
        self.name = "nse_derivatives"
        self.options.col_prefix = "FO-"
        self.rescale_options.turnover_scale = math.pow(10, 7)
        self.options.unzip_file = False
        __base_url = "https://archives.nseindia.com/content/fo/"
        self.options.url_template = Template(__base_url + "$download_filename")
        self.options.output_path_template = Template(
            "$DataBaseDir/$RawDataDir/$NseDerivativesDataDir/$download_filename"
        )
        self.options.primary_data_path_template = Template(
            "$DataBaseDir/$RawDataDir/$NseDerivativesDataDir/$download_filename"
        )

    def get_filenames(self, for_date):
        # __formatted_date = for_date.strftime('%d%b%Y').upper()
        __formatted_date = for_date.strftime("%d%m%Y")
        return {
            "download_filename": f"NSE_FO_bhavcopy_{__formatted_date}.csv",
            "primary_data_filename": f"NSE_FO_bhavcopy_{__formatted_date}.csv",
        }

    def get_column_name_mappings(self):
        return {
            "TckrSymb": BaseColumns.Identifier,
            "XpryDt": DerivativesBaseColumns.ExpiryDate,
            "OptnTp": DerivativesBaseColumns.OptionType,
            "PrvsClsgPric": BaseColumns.PreviousClose,
            "TtlTradgVol": BaseColumns.Volume,
            "TtlTrfVal": BaseColumns.Turnover,
            "OpnPric": BaseColumns.Open,
            "HghPric": BaseColumns.High,
            "LwPric": BaseColumns.Low,
            "ClsPric": BaseColumns.Close,
            "OpnIntrst": DerivativesBaseColumns.OpenInterest,
            "PctgChngInOpnIntrst": DerivativesBaseColumns.OiChangePct,
            "StrkPric": DerivativesBaseColumns.StrikePrice,
            "FinInstrmNm": DerivativesBaseColumns.InstrumentType,
        }


class NseEquityFuturesDataReader(NseDerivatiesReader):
    def __init__(self):
        super().__init__()
        self.name = "nse_futstk"
        self.options.col_prefix = "Futures-"
        self.set_filter(InstrumentTypeFilter("FUTSTK"))


class NseDerivatiesOldReader(NseDerivatiesReaderBase):
    def __init__(self):
        super().__init__()
        self.name = "Derivatives"
        self.options.col_prefix = "FO-"
        self.rescale_options.turnover_scale = math.pow(10, 7)
        __base_url = "https://archives.nseindia.com/content/historical/DERIVATIVES/"
        self.options.url_template = Template(
            __base_url + "$year/$month/$download_filename"
        )
        self.options.output_path_template = Template(
            "$DataBaseDir/$RawDataDir/$NseDerivativesDataDir/$download_filename"
        )
        self.options.unzip_path_template = Template(
            "$DataBaseDir/$RawDataDir/$NseDerivativesDataDir/$download_filename_wo_ext"
        )
        self.options.primary_data_path_template = Template(
            "$DataBaseDir/$RawDataDir/$NseDerivativesDataDir/$download_filename_wo_ext/$primary_data_filename"
        )

    def get_filenames(self, for_date):
        __formatted_date = for_date.strftime("%d%b%Y").upper()
        return {
            "download_filename_wo_ext": f"fo{__formatted_date}bhav",
            "download_filename": f"fo{__formatted_date}bhav.csv.zip",
            "primary_data_filename": f"fo{__formatted_date}bhav.csv",
        }

    def get_column_name_mappings(self):
        return {
            "SYMBOL": BaseColumns.Identifier,
            "EXPIRY_DT": DerivativesBaseColumns.ExpiryDate,
            "OPTION_TYP": DerivativesBaseColumns.OptionType,
            "PrvsClsgPric": BaseColumns.PreviousClose,
            "TtlTradgVol": BaseColumns.Turnover,
            "OPEN": BaseColumns.Open,
            "HIGH": BaseColumns.High,
            "LOW": BaseColumns.Low,
            "CLOSE": BaseColumns.Close,
            "OpnIntrst": DerivativesBaseColumns.OpenInterest,
            "PctgChngInOpnIntrst": DerivativesBaseColumns.OiChangePct,
            "STRIKE_PR": DerivativesBaseColumns.StrikePrice,
        }


class NseIndexFuturesDataReader(NseDerivatiesReader):
    def __init__(self):
        super().__init__()
        self.name = "nse_futidx"
        self.options.col_prefix = "Futures-"
        self.set_filter(InstrumentTypeFilter("FUTIDX"))
        
    
    def sanitize_data(self, data: pd.DataFrame):
        mapping = {
            "NIFTY": "Nifty 50",
            "BANKNIFTY": "Nifty Bank",
            "FINNIFTY": "Nifty Financial Services",
        }
        data.loc[data.index, BaseColumns.Identifier] = data[BaseColumns.Identifier].replace(mapping)
        return data


class NseIndexOptionsDataReader(NseDerivatiesReader):
    def __init__(self):
        super().__init__()
        self.name = "nse_optidx"
        self.options.col_prefix = "Option-"
        self.set_filter(InstrumentTypeFilter("OPTIDX"))
    
    def sanitize_data(self, data: pd.DataFrame):
        mapping = {
            "NIFTY": "Nifty 50",
            "BANKNIFTY": "Nifty Bank",
            "FINNIFTY": "Nifty Financial Services",
        }
        data.loc[data.index, BaseColumns.Identifier] = data[BaseColumns.Identifier].replace(mapping)
        return data


class NseEquityOptionsDataReader(NseDerivatiesReader):
    def __init__(self):
        super().__init__()
        self.name = "nse_optstk"
        self.options.col_prefix = "Option-"
        self.set_filter(InstrumentTypeFilter("OPTSTK"))


class NseIndicesManualDataReader(DateRangeSourceDataReader):
    output_dir_template = Template("")
    filename_template = Template("$ReaderName.csv")
    data_dir_template = Template("$DataBaseDir/$ManualDataDir")

    def __init__(self):
        super().__init__()
        self.name = "nse_indices"
        self.options.col_prefix = "index-"
        self.options.data_availibility = ReaderDataAvailibility(from_date=date(1990, 7, 3), till_date=date(2012, 12, 31))

    def read(self, criteria: DateRangeCriteria):
        data_file = os.path.join(
            self.data_dir_template.substitute(**EnvironmentSettings.Paths),
            self.filename_template.substitute(**{"ReaderName": "Indices"}),
        )

        if os.path.exists(data_file):
            data = pd.read_csv(data_file)
            data[BaseColumns.Date] = pd.to_datetime(
                data[BaseColumns.Date]
            )
            data = data[
                data[BaseColumns.Date].dt.date.between(criteria.from_date, criteria.to_date)
            ]
            
            return self.post_read_data(data)
        else:
            return pd.DataFrame()


class NseIndicesReader(ChainedDataReader):
    def __init__(self):
        super().__init__(NseIndicesManualDataReader())
        self.name = "nse_indices"
        self.options.col_prefix = "index-"
        self.new_reader = DateRangeDataReaderWrapper(NseIndicesNewReader())
    
    def read_data(self, criteria: ReaderDateCriteria):
        return self.new_reader.read(criteria)
    
    def has_data(self, criteria: ReaderDateCriteria):
        return self.new_reader.has_data(criteria)


import pytest
import pandas as pd
import os
import shutil

from helper import check_col_values, setup, check_base_cols_present, Presets
from markets_insights.core.core import IdentifierFilter
from markets_insights.core.column_definition import BaseColumns
from markets_insights.core.environment import Environment, EnvironmentSettings

from markets_insights.datareader.data_reader import DateRangeCriteria, NseIndicesReader
from markets_insights.dataprocess.data_processor import (
    HistoricalDataProcessOptions,
    HistoricalDataProcessor,
    HistoricalDataset,
)

old_path = EnvironmentSettings.Paths["DataBaseDir"]

def clear_output_folder():
    data_folder_path = os.path.join(os.getcwd(), 'data')

    def clean_data_folder():
        shutil.rmtree(data_folder_path)

    if os.path.isdir(data_folder_path):
        clean_data_folder()
    else:
        os.mkdir(data_folder_path)

    Environment.setup(cache_data_base_path=data_folder_path)

setup()

@pytest.mark.parametrize("identifier,unique_ids_count",
    [
        ("Nifty 50", 1),
        (None, 108)        
    ]
)
def test_historical_data_processor_caching_without_filter(identifier: str, unique_ids_count: int):
    processor = HistoricalDataProcessor(HistoricalDataProcessOptions(include_annual_data=False, include_monthly_data=False))
    reader = NseIndicesReader()
    
    if identifier:
        reader.set_filter(IdentifierFilter(identifier))

    result: HistoricalDataset = processor.process(
        reader,
        DateRangeCriteria(Presets.dates.dec_start, Presets.dates.dec_end),
    )
    unique_ids = result.get_daily_data()[BaseColumns.Identifier].unique()
    assert len(unique_ids) == unique_ids_count

Environment.setup(cache_data_base_path=old_path)


from markets_insights.core.column_definition import (
    BaseColumns,
    CalculatedColumns,
    CalculatedColumnsBase,
    DerivativesBaseColumns,
)
from markets_insights.core.settings import MarketDaysSettings
from markets_insights.core.environment import EnvironmentSettings
from datetime import date
import time
import pandas as pd
from dateutil.relativedelta import relativedelta
import calendar

Weekends = ["Saturday", "Sunday"]


class TypeHelper:
    def get_class_props(c_type):
        return [
            x
            for x in dir(c_type)
            if x.startswith("__") == False and type(getattr(c_type, x)) != "function"
        ]

    def get_class_static_values(c_type):
        return [
            getattr(c_type, x)
            for x in dir(c_type)
            if x.startswith("__") == False and type(getattr(c_type, x)) != "function"
        ]

    def get_class_props_for_value(c_type, val):
        props = TypeHelper.get_class_props(c_type)
        return [prop for prop in props if val in getattr(c_type, prop)]


class InstrumentationType:
    Info = 1
    Trace = 2
    Debug = 4


InstrumentationLevel = EnvironmentSettings.Development["InstrumentationLevel"]


class Instrumentation:
    def debug(message):
        if (
            InstrumentationLevel & InstrumentationType.Debug
            == InstrumentationType.Debug
        ):
            print(message)

    def info(message):
        if InstrumentationLevel & InstrumentationType.Info == InstrumentationType.Info:
            print(message)

    def startTracing(name):
        if (
            InstrumentationLevel & InstrumentationType.Trace
            == InstrumentationType.Trace
        ):
            return Trace(name)
        else:
            return NoTrace(name)

    def trace(name):
        def decorator(function):
            def wrapper(*args, **kwargs):
                tracer = Instrumentation.startTracing(name)
                result = function(*args, **kwargs)
                tracer.endTracing()
                return result

            return wrapper

        return decorator


class Trace:
    def __init__(self, name):
        self.startTime = time.time()
        self.name = name

    def endTracing(self):
        print(f"{self.name} took {round(time.time() - self.startTime)} seconds")


class NoTrace:
    def __init__(self, name):
        self.name = name

    def endTracing(self):
        return


class MarketDaysHelper:
    def is_open_for_day(for_date):
        day = for_date.strftime("%A")
        if hasattr(for_date, "date"):
            for_date = for_date.date()
        return not (day in Weekends or for_date in MarketDaysSettings.MarketHolidays)

    def get_this_or_next_market_day(for_date):
        current_date = for_date
        i = 1
        while i < 5:
            if not MarketDaysHelper.is_open_for_day(current_date):
                current_date = current_date + pd.Timedelta(days=1)
            else:
                return current_date
            i = i + 1
        return current_date

    def get_next_market_day(for_date):
        current_date = for_date + pd.Timedelta(days=1)
        i = 1
        while i < 5:
            if not MarketDaysHelper.is_open_for_day(current_date):
                current_date = current_date + pd.Timedelta(days=1)
            else:
                return current_date
            i = i + 1
        return current_date

    def get_this_or_previous_market_day(for_date):
        current_date = for_date
        i = 1
        while i < 5:
            if not MarketDaysHelper.is_open_for_day(current_date):
                current_date = current_date - pd.Timedelta(days=1)
            else:
                return current_date
            i = i + 1
        return current_date

    def get_previous_market_day(for_date):
        current_date = for_date - pd.Timedelta(days=1)
        i = 1
        while i < 5:
            if not MarketDaysHelper.is_open_for_day(current_date):
                current_date = current_date - pd.Timedelta(days=1)
            else:
                return current_date
            i = i + 1
        return current_date

    def get_days_list_for_range(from_date, to_date):
        return pd.date_range(from_date, to_date, freq="B").tolist()

    def get_monthly_expiry_dates(no_of_recent_expiries):
        expiries = []
        today = date.today()
        for i in range(1, no_of_recent_expiries):
            cur_month = date(today.year, today.month, 1) - relativedelta(months=i)
            monthly_expiry = MarketDaysHelper.get_this_or_previous_market_day(
                MarketDaysHelper.get_last_thursday(cur_month.year, cur_month.month)
            )
            expiries.append(monthly_expiry)
        return expiries

    def get_last_thursday(year, month):
        cal = calendar.Calendar(firstweekday=calendar.MONDAY)
        thursday_number = -1
        monthcal = cal.monthdatescalendar(year, month)
        last_thursday = [
            day
            for week in monthcal
            for day in week
            if day.weekday() == calendar.THURSDAY and day.month == month
        ][thursday_number]
        return last_thursday


class FilterCriteria:
    _col_to_filter: str = None
    _condition: str = None
    _condition_value = None

    def __init__(
        self, col_to_filter: str = None, condition: str = None, condition_value = None
    ):
        self._col_to_filter = col_to_filter
        self._condition = condition
        self._condition_value = condition_value

    def __str__(self) -> str:
        return self.get_query()

    def get_query(self) -> str:
        return f"`{self._col_to_filter}` {self._condition} {self._condition_value}"


class FilterBase:
    _filter_criterias: [FilterCriteria]

    def __init__(self):
        self._filter_criterias = []

    def add_criteria(self, criteria: FilterCriteria):
        self._filter_criterias.append(criteria)

    def __str__(self) -> str:
        return f"{self.get_query()}"

    def get_query(self) -> str:
        return " & ".join([f"{k.get_query()}" for k in self._filter_criterias])

    def __and__(self, other: object):
        new_filter = FilterBase()
        for criteria in self._filter_criterias:
            new_filter.add_criteria(criteria)
        for criteria in other._filter_criterias:
            new_filter.add_criteria(criteria)
        return new_filter


class IdentifierFilter(FilterBase):
    def __init__(self, identifier: str):
        super().__init__()
        self.add_criteria(
            FilterCriteria(
                col_to_filter=BaseColumns.Identifier,
                condition="==",
                condition_value=f"'{identifier}'",
            )
        )


class InstrumentTypeFilter(FilterBase):
    def __init__(self, identifier: str):
        super().__init__()
        self.add_criteria(
            FilterCriteria(
                col_to_filter=DerivativesBaseColumns.InstrumentType,
                condition="==",
                condition_value=f"'{identifier}'",
            )
        )


class DateFilter(FilterBase):
    def __init__(self, for_date: date = None):
        super().__init__()
        self.add_criteria(
            FilterCriteria(
                col_to_filter=BaseColumns.Date,
                condition="==",
                condition_value=f"'{for_date}'",
            )
        )


class DateRangeFilter(FilterBase):
    def __init__(self, from_date: date = None, to_date: date = None):
        super().__init__()
        if from_date is not None:
            self.add_criteria(
                FilterCriteria(
                    col_to_filter=BaseColumns.Date,
                    condition=">=",
                    condition_value=f"'{from_date}'",
                )
            )

        if to_date is not None:
            self.add_criteria(
                FilterCriteria(
                    col_to_filter=BaseColumns.Date,
                    condition="<=",
                    condition_value=f"'{to_date}'",
                )
            )


class DatePartsFilter(FilterBase):
    def __init__(self, month_no: int = None, year: int = None):
        super().__init__()
        if month_no is not None:
            self.add_criteria(
                FilterCriteria(
                    col_to_filter=CalculatedColumns.MonthNo,
                    condition="==",
                    condition_value=month_no,
                )
            )
        if year is not None:
            self.add_criteria(
                FilterCriteria(
                    col_to_filter=CalculatedColumns.Year,
                    condition="==",
                    condition_value=year,
                )
            )


class ExpiryDateFilter(FilterBase):
    def __init__(self, date: str):
        super().__init__()
        self.add_criteria(
            FilterCriteria(
                col_to_filter=DerivativesBaseColumns.ExpiryDate,
                condition="==",
                condition_value=f"'{date}'",
            )
        )


class FlagFilter(FilterBase):
    def __init__(self, flag_column_name: str):
        super().__init__()
        self.add_criteria(
            FilterCriteria(
                col_to_filter=flag_column_name, condition="==", condition_value=True,
            )
        )


import os
import sys

code_dir = os.path.join(os.path.abspath("."), "src")
os.chdir(code_dir)
sys.path.append(code_dir)

import datetime
from markets_insights.datareader.data_reader import BhavCopyReader, DateRangeCriteria, NseIndicesReader
from markets_insights.dataprocess.data_processor import HistoricalDataProcessor, MultiDataCalculationPipelines, CalculationPipelineBuilder, HistoricalDataProcessOptions
from markets_insights.calculations.base import DatePartsCalculationWorker

reader = BhavCopyReader()
histDataProcessor = HistoricalDataProcessor()

# Fetch the data
year_start = datetime.date(2023, 1, 1)
to_date = datetime.date(2023, 12, 31)
result = histDataProcessor.process(reader, DateRangeCriteria(year_start, to_date))

# # Prepare calculation pipeline
# pipelines = MultiDataCalculationPipelines()
# pipelines.set_item('rsi', CalculationPipelineBuilder.create_rsi_calculation_pipeline())
# histDataProcessor.set_calculation_pipelines(pipelines)

# # Run the pipeline
# histDataProcessor.run_calculation_pipelines()


from markets_insights.core.environment import EnvironmentSettings
import os
from string import Template
import pandas as pd
from markets_insights.datareader.data_reader import DataReader, DateRangeSourceDataReader, DateRangeDataReaderWrapper, DateRangeCriteria
from markets_insights.core.column_definition import (
    BaseColumns,
    BasePriceColumns,
    CalculatedColumns,
    PeriodAggregateColumnTemplate,
    AggregationPeriods,
)
from markets_insights.core.core import MarketDaysHelper, Instrumentation, TypeHelper
from markets_insights.core.column_definition import BaseColumns
from markets_insights.calculations.base import (
    CalculationWindow,
    HighestPriceInNextNDaysCalculationWorker,
    LowestPriceInNextNDaysCalculationWorker,
)
from markets_insights.calculations.base import (
    CalculationWorker,
    CalculationPipeline,
    SmaCalculationWorker,
    RsiCalculationWorker,
    BollingerBandCalculationWorker,
    StochRsiCalculationWorker,
    ColumnValueCrossedAboveFlagWorker,
    ColumnValueCrossedBelowFlagWorker,
    PriceCrossedAboveColumnValueFlagWorker,
    PriceCrossedBelowColumnValueFlagWorker,
    StdDevCalculationWorker,
)

from datetime import date
import glob
from typing import Dict


class DataProcessor:
    output_dir_template = Template("")
    filename_template = Template("$ReaderName.csv")
    manual_data_dir_template = Template("$ManualDataPath")
    options: dict

    def process(self, reader: DataReader):
        return Exception("Not implemented!")

    def remove_unnamed_columns(self, data: pd.DataFrame) -> pd.DataFrame:
        unnamed_cols = [col for col in data.columns if "Unnamed" in col]
        if len(unnamed_cols):
            data.drop(columns=unnamed_cols, inplace=True)


class CalculationPipelineBuilder:
    def create_pipeline_for_worker(worker: CalculationWorker):
        pipeline = CalculationPipeline()
        pipeline.add_calculation_worker(worker)
        return pipeline

    def create_pipeline_for_workers(workers: list[CalculationWorker]):
        pipeline = CalculationPipeline()
        for worker in workers:
            pipeline.add_calculation_worker(worker)
        return pipeline

    def create_bb_calculation_pipeline(windows=[200], deviations=[2, 3]):
        pipeline = CalculationPipeline()
        for window in windows:
            pipeline.add_calculation_worker(StdDevCalculationWorker(time_window=window))
            for deviation in deviations:
                worker = BollingerBandCalculationWorker(window, deviation)
                pipeline.add_calculation_worker(worker)
                pipeline.add_calculation_worker(
                    PriceCrossedBelowColumnValueFlagWorker(worker.get_columns()[0])
                )
                pipeline.add_calculation_worker(
                    PriceCrossedAboveColumnValueFlagWorker(worker.get_columns()[1])
                )
        return pipeline

    def create_sma_calculation_pipeline(windows=[50, 100, 200]):
        pipeline = CalculationPipeline()
        for window in windows:
            worker = SmaCalculationWorker(window)
            pipeline.add_calculation_worker(worker)
            pipeline.add_calculation_worker(
                PriceCrossedBelowColumnValueFlagWorker(worker.get_column())
            )
            pipeline.add_calculation_worker(
                PriceCrossedAboveColumnValueFlagWorker(worker.get_column())
            )
        return pipeline

    def create_rsi_calculation_pipeline(
        crossing_above_flag_value=75, crossing_below_flag_value=30, window=14
    ):
        pipeline = CalculationPipeline()
        pipeline.add_calculation_worker(RsiCalculationWorker(window))
        if crossing_above_flag_value is not None:
            pipeline.add_calculation_worker(
                ColumnValueCrossedAboveFlagWorker(
                    CalculatedColumns.RelativeStrengthIndex, crossing_above_flag_value
                )
            )
        if crossing_below_flag_value is not None:
            pipeline.add_calculation_worker(
                ColumnValueCrossedBelowFlagWorker(
                    CalculatedColumns.RelativeStrengthIndex, crossing_below_flag_value
                )
            )
        return pipeline

    def create_stoch_rsi_calculation_pipeline(
        crossing_above_flag_value=80, crossing_below_flag_value=20, window=14
    ):
        pipeline = CalculationPipeline()
        pipeline.add_calculation_worker(StochRsiCalculationWorker(window))
        if crossing_above_flag_value is not None:
            pipeline.add_calculation_worker(
                ColumnValueCrossedAboveFlagWorker(
                    CalculatedColumns.StochRsi_K, crossing_above_flag_value
                )
            )
            pipeline.add_calculation_worker(
                ColumnValueCrossedAboveFlagWorker(
                    CalculatedColumns.StochRsi_D, crossing_above_flag_value
                )
            )
        if crossing_below_flag_value is not None:
            pipeline.add_calculation_worker(
                ColumnValueCrossedBelowFlagWorker(
                    CalculatedColumns.StochRsi_K, crossing_below_flag_value
                )
            )
            pipeline.add_calculation_worker(
                ColumnValueCrossedBelowFlagWorker(
                    CalculatedColumns.StochRsi_D, crossing_below_flag_value
                )
            )
        return pipeline

    def create_forward_looking_price_fall_pipeline(n_days_list):
        forward_looking_lowest_price_pipeline = CalculationPipeline()
        for n in n_days_list:
            forward_looking_lowest_price_pipeline.add_calculation_worker(
                LowestPriceInNextNDaysCalculationWorker(n)
            )
        return forward_looking_lowest_price_pipeline

    def create_forward_looking_price_rise_pipeline(n_days_list):
        forward_looking_highest_price_pipeline = CalculationPipeline()
        for n in n_days_list:
            forward_looking_highest_price_pipeline.add_calculation_worker(
                HighestPriceInNextNDaysCalculationWorker(n)
            )
        return forward_looking_highest_price_pipeline


class MultiDataCalculationPipelines:
    def __init__(self):
        self._store: Dict[str, CalculationPipeline] = {}

    def set_item(self, k: str, v: CalculationPipeline) -> None:
        self._store[k] = v

    def get_item(self, k: str) -> CalculationPipeline:
        return self._store[k]

    def run(self, data):
        for key in self._store:
            result = self._store[key].run(data)
            if result is not None:
                data = result
        return data

    def get_calculation_window(self) -> CalculationWindow:
        return CalculationWindow.load_from_list([self._store[key].get_calculation_window() for key in self._store])

class HistoricalDataset:
    _daily: pd.DataFrame = None
    _identifier_grouped: pd.core.groupby.DataFrameGroupBy = None
    _monthly: pd.DataFrame = None
    _annual: pd.DataFrame = None

    def create_identifier_grouped(self):
        self._identifier_grouped = self.create_grouped_data(BaseColumns.Identifier)
        return self

    def set_daily_data(self, daily: pd.DataFrame):
        self._daily = daily
        return self

    def set_monthly_data(self, monthly: pd.DataFrame):
        self._monthly = monthly
        return self

    def set_annual_data(self, annual: pd.DataFrame):
        self._annual = annual
        return self

    def create_grouped_data(self, columns):
        return self._daily.groupby(columns)

    def get_identifier_grouped(self):
        return self._identifier_grouped

    def get_daily_data(self) -> pd.DataFrame:
        return self._daily

    def get_monthly_data(self) -> pd.DataFrame:
        return self._monthly

    def get_annual_data(self) -> pd.DataFrame:
        return self._annual

class HistoricalDataProcessOptions:
    def __init__(
        self, include_monthly_data: bool = True, include_annual_data: bool = True
    ):
        self.include_monthly_data = include_monthly_data
        self.include_annual_data = include_annual_data


class HistoricalDataProcessor(DataProcessor):
    options: HistoricalDataProcessOptions

    def __init__(
        self, options: HistoricalDataProcessOptions = HistoricalDataProcessOptions()
    ):
        self.output_dir_template = Template(
            "$DataBaseDir/$ProcessedDataDir/$HistoricalDataDir/"
        )
        self.monthly_group_data_filename = Template("$ReaderName-$MonthlySuffix.csv")
        self.annual_group_data_filename = Template("$ReaderName-$AnnualSuffix.csv")
        self.historic_highs_reset_days = 60
        self.dataset: HistoricalDataset
        self.calculation_pipelines: MultiDataCalculationPipelines
        self.options = options

    def set_calculation_pipelines(self, pipelines):
        self.calculation_pipelines = pipelines

    def run_calculation_pipelines(self):
        daily_data = self.calculation_pipelines.run(self.dataset.get_daily_data())
        if daily_data is not None:
            self.dataset.set_daily_data(daily_data)


    @Instrumentation.trace(name="HistoricalDataProcessor.process")
    def process(self, reader: DataReader, criteria: DateRangeCriteria) -> HistoricalDataset:
        from_date = MarketDaysHelper.get_this_or_next_market_day(criteria.from_date)
        to_date = MarketDaysHelper.get_this_or_previous_market_day(criteria.to_date)

        daily_data = pd.DataFrame(self.get_data(reader, from_date, to_date).drop_duplicates())

        if not daily_data.empty:
            if reader.filter:
                daily_data = pd.DataFrame(daily_data.query(reader.filter.get_query()))

            self.dataset = HistoricalDataset()
            self.dataset.set_daily_data(daily_data)
            self.dataset.create_identifier_grouped()

            output_path = self.output_dir_template.substitute(**EnvironmentSettings.Paths)

            if self.options.include_monthly_data == True:
                self.dataset.set_monthly_data(self.add_monthly_growth_calc(daily_data))
                monthly_data_filename = self.monthly_group_data_filename.substitute(
                    {**EnvironmentSettings.Paths, "ReaderName": reader.name}
                )
                self.dataset.get_monthly_data().to_csv(output_path + monthly_data_filename)

            if self.options.include_annual_data == True:
                self.dataset.set_annual_data(self.add_yearly_growth_calc(daily_data))
                annual_data_filename = self.annual_group_data_filename.substitute(
                    {**EnvironmentSettings.Paths, "ReaderName": reader.name}
                )
                self.dataset.get_annual_data().to_csv(output_path + annual_data_filename)
        else:
            self.dataset = HistoricalDataset()
            self.dataset.set_daily_data(daily_data)

        return self.dataset
    

    def add_periodic_growth_calc(
        self, processed_data: pd.DataFrame, period: str
    ) -> pd.DataFrame:
        Instrumentation.debug(f"Started periodic calculation for {period}")

        periodic_grouped = processed_data.groupby([BaseColumns.Identifier, period])

        processed_data[
            PeriodAggregateColumnTemplate.substitute(
                {"period": period, "col_name": BaseColumns.Open}
            )
        ] = periodic_grouped[BaseColumns.Open].transform(lambda x: x.iloc[0])
        processed_data[
            PeriodAggregateColumnTemplate.substitute(
                {"period": period, "col_name": BaseColumns.Close}
            )
        ] = periodic_grouped[BaseColumns.Close].transform(lambda x: x.iloc[-1])
        processed_data[
            PeriodAggregateColumnTemplate.substitute(
                {"period": period, "col_name": BaseColumns.Low}
            )
        ] = periodic_grouped[BaseColumns.Low].transform("min")
        processed_data[
            PeriodAggregateColumnTemplate.substitute(
                {"period": period, "col_name": BaseColumns.High}
            )
        ] = periodic_grouped[BaseColumns.High].transform("max")

        processed_data[
            PeriodAggregateColumnTemplate.substitute(
                {"period": period, "col_name": BaseColumns.Volume}
            )
        ] = periodic_grouped[BaseColumns.Volume].transform("sum")
        processed_data[
            PeriodAggregateColumnTemplate.substitute(
                {"period": period, "col_name": BaseColumns.Turnover}
            )
        ] = periodic_grouped[BaseColumns.Turnover].transform("sum")

        periodic_data = periodic_grouped.last().reset_index()

        aggregated_cols = [
            BaseColumns.Volume,
            BaseColumns.Turnover,
        ] + TypeHelper.get_class_static_values(BasePriceColumns)

        periodic_data = periodic_data[
            [BaseColumns.Identifier, BaseColumns.Date, period]
            + [
                PeriodAggregateColumnTemplate.substitute(
                    {"period": period, "col_name": col_name}
                )
                for col_name in aggregated_cols
            ]
        ]

        cols = {}
        for col_name in aggregated_cols:
            cols[
                PeriodAggregateColumnTemplate.substitute(
                    {"period": period, "col_name": col_name}
                )
            ] = col_name
        periodic_data.rename(columns=cols, inplace=True)

        return periodic_data

    @Instrumentation.trace(name="HistoricalDataProcessor.add_yearly_growth_calc")
    def add_yearly_growth_calc(self, processed_data):
        processed_data[CalculatedColumns.Year] = processed_data[
            BaseColumns.Date
        ].dt.year
        return self.add_periodic_growth_calc(processed_data, CalculatedColumns.Year)

    @Instrumentation.trace(name="HistoricalDataProcessor.add_monthly_growth_calc")
    def add_monthly_growth_calc(self, processed_data):
        processed_data[CalculatedColumns.Month] = processed_data[
            BaseColumns.Date
        ].dt.strftime("%Y-%m")
        return self.add_periodic_growth_calc(processed_data, CalculatedColumns.Month)

    def rename_columns(self, reader: DataReader, historical_data: pd.DataFrame):
        column_name_mappings = reader.get_column_name_mappings()
        if column_name_mappings is not None:
            historical_data.rename(columns=column_name_mappings, inplace=True)

    @Instrumentation.trace(name="HistoricalDataProcessor.get_data")
    def get_data(
        self, reader: DataReader, from_date: date, to_date: date
    ):
        Instrumentation.debug("Started to read data")
        output_file = os.path.join(
            self.output_dir_template.substitute(**EnvironmentSettings.Paths),
            self.filename_template.substitute(**{"ReaderName": reader.name}),
        )

        if isinstance(reader, DateRangeSourceDataReader):
            dateRangeReader = reader
        else:
            dateRangeReader = DateRangeDataReaderWrapper(reader)

        save_to_file = False

        Instrumentation.info(f"Reading data from {from_date} to {to_date}")
        read_data = dateRangeReader.read(DateRangeCriteria(from_date, to_date))
        
        if not read_data.empty:
            if save_to_file == True:
                Instrumentation.debug(f"Saving data to file: {output_file}")
                read_data.to_csv(output_file, index=False)

            try:
                read_data[BaseColumns.Date] = read_data[BaseColumns.Date].str.replace(' 00:00:00', '')
            except:
                pass

            read_data[BaseColumns.Date] = pd.to_datetime(read_data[BaseColumns.Date], format="%Y-%m-%d")
            
            read_data = read_data[
                read_data[BaseColumns.Date].dt.date.between(from_date, to_date)
            ]
        
        return read_data



import os


class EnvironmentSettings:
    Paths = {
        "DataBaseDir": "../../_data",
        "RawDataDir": "raw",
        "ProcessedDataDir": "processed",
        "HistoricalDataDir": "historical",
        "MonthlySuffix": "monthly",
        "AnnualSuffix": "annual",
        "BhavDataDir": "bhavcopy",
        "NseDerivativesDataDir": "nse_derivatives",
        "NseIndicesDataDir": "nse_indices",
        "ManualDataDir": "manual_data",
        "ManualDataPath": "../manual_data",
    }
    Development = {"InstrumentationLevel": 1 | 2 | 4}


class Environment:
    def setup(cache_data_base_path: str = None):
        if cache_data_base_path is not None:
            EnvironmentSettings.Paths["DataBaseDir"] = cache_data_base_path

        Environment._setup_cache_dirs()

    def _setup_cache_dirs():
        env_paths = EnvironmentSettings.Paths

        cur_path = f"{env_paths['DataBaseDir']}"
        if not os.path.exists(cur_path):
            os.mkdir(cur_path)

        for folder in ["RawDataDir", "ProcessedDataDir", "ManualDataDir"]:
            cur_path = f"{env_paths['DataBaseDir']}/{env_paths[folder]}"
            if not os.path.exists(cur_path):
                os.mkdir(cur_path)

        for folder in ["BhavDataDir", "NseIndicesDataDir", "NseDerivativesDataDir"]:
            cur_path = f"{env_paths['DataBaseDir']}/{env_paths['RawDataDir']}/{env_paths[folder]}"
            if not os.path.exists(cur_path):
                os.mkdir(cur_path)

        cur_path = f"{env_paths['DataBaseDir']}/{env_paths['ProcessedDataDir']}/{env_paths['HistoricalDataDir']}"
        if not os.path.exists(cur_path):
            os.mkdir(cur_path)
